import json
import typing as t
from functools import reduce
from urllib.parse import parse_qsl, urlencode

from channels.db import database_sync_to_async as db
from channels.layers import BaseChannelLayer
from django.contrib.auth.models import AbstractBaseUser, AnonymousUser

from .component import Component, MessagePayload
from .utils import filter_parameters


class ComponentRepository:
    user: AnonymousUser | AbstractBaseUser

    def __init__(
        self,
        user: AnonymousUser | AbstractBaseUser | None = None,
        params: dict[str, t.Any] | None = None,
        channel_name: str | None = None,
        channel_layer: BaseChannelLayer | None = None,
    ):
        self.params = params or {}
        self.channel_name = channel_name
        self.channel_layer = channel_layer
        self.user = user or AnonymousUser()
        self.components: dict[str, Component] = {}

    @staticmethod
    def extract_params(qs: str):
        return {
            key: json.loads(value) if key.endswith(".json") else value
            for key, value in parse_qsl(qs)
        }

    def set_query_string(self, qs: str):
        params = self.extract_params(qs)

        # remove old keys
        for key in list(self.params.keys()):
            if key not in params:
                self.params.pop(key)

        self.params.update(params)

    def get_query_string(self) -> str:
        return urlencode(
            {
                key: json.dumps(value) if key.endswith(".json") else value
                for key, value in self.params.items()
            }
        )

    def get(self, component_id: str) -> Component | None:
        return self.components.get(component_id)

    def build(
        self, name: str, state: MessagePayload, parent_id: str | None = None
    ) -> tuple[Component, bool]:
        if (component_id := state.get("id")) and (
            component := self.components.get(component_id)
        ):
            # override with the passed state but preserve the rest of the state
            for key, value in state.items():
                setattr(component, key, value)
            component.reactor.parent_id = parent_id
            return component, False
        else:
            component = Component._build(
                name,
                state,
                params=self.params,
                user=self.user,
                channel_name=self.channel_name,
                channel_layer=self.channel_layer,
                parent_id=parent_id,
            )
            return self.register_component(component), True

    async def join(
        self, name: str, state: MessagePayload, parent_id: str | None = None
    ) -> tuple[Component, bool]:
        component, created = await db(self.build)(
            name, state, parent_id=parent_id
        )
        if created:
            await component.joined()
        return component, created

    def register_component(self, component: Component):
        self.components[component.id] = component
        return component

    def remove(self, id):
        self.components.pop(id, None)

    async def dispatch_event(self, id, command, kwargs):
        assert not command.startswith("_")
        component = self.components[id]
        handler = getattr(component, command)
        await handler(**filter_parameters(handler, kwargs))
        return component

    def components_subscribed_to(self, channel):
        # XXX: There is a list() here because the dict can change size during
        # iteration
        for component in list(self.components.values()):
            if channel in component._subscriptions:
                yield component

    @property
    def subscriptions(self):
        return reduce(
            lambda a, b: a.union(b),
            (
                component._subscriptions
                for component in list(self.components.values())
            ),
            set(),
        )



import logging
import typing as t

from django.apps import apps
from django.db import models
from django.db.models.signals import m2m_changed, post_save, pre_delete
from django.dispatch import Signal

from . import serializer
from .schemas import ModelAction
from .settings import AUTO_BROADCAST
from .utils import send_to

__all__ = []

log = logging.getLogger("reactor")


senders = (
    [
        apps.get_model(app_label, model_name)
        for app_label, model_name in AUTO_BROADCAST.senders
    ]
    if AUTO_BROADCAST.senders
    else [None]
)


def receiver(signal: Signal, *, is_active: bool):
    def _decorator(f):
        if is_active:
            for sender in senders:
                signal.connect(f, sender=sender)
        return f

    return _decorator


@receiver(
    post_save,
    is_active=AUTO_BROADCAST.model
    or AUTO_BROADCAST.model_pk
    or AUTO_BROADCAST.related,
)
def broadcast_post_save(sender, instance, created=False, **kwargs):
    name = sender._meta.model_name
    encoded_instance = serializer.encode(instance)
    action = ModelAction.CREATED if created else ModelAction.UPDATED
    if AUTO_BROADCAST.model:
        notify_mutation([name], action, encoded_instance)

    if instance.pk is not None:
        if AUTO_BROADCAST.model_pk:
            notify_mutation(
                [f"{name}.{instance.pk}"],
                action,
                encoded_instance,
            )
        if AUTO_BROADCAST.related:
            broadcast_related(
                sender,
                action,
                instance,
                encoded_instance,
            )


@receiver(
    pre_delete,
    is_active=AUTO_BROADCAST.model
    or AUTO_BROADCAST.model_pk
    or AUTO_BROADCAST.related,
)
def broadcast_pre_delete(sender, instance, **kwargs):
    name = sender._meta.model_name
    encoded_instance = serializer.encode(instance)
    if AUTO_BROADCAST.model:
        notify_mutation([name], ModelAction.DELETED, encoded_instance)

    if instance.pk is not None:
        if AUTO_BROADCAST.model_pk:
            notify_mutation(
                [f"{name}.{instance.pk}"],
                ModelAction.DELETED,
                encoded_instance,
            )
        if AUTO_BROADCAST.related:
            broadcast_related(
                sender,
                ModelAction.DELETED,
                instance,
                encoded_instance,
            )


def broadcast_related(sender, action: ModelAction, instance, encoded_instance):
    for field in get_related_fields(sender):
        if field["is_m2m"]:
            fk_ids = getattr(instance, field["name"]).values_list(
                "id", flat=True
            )
        else:
            fk_ids = filter(None, [getattr(instance, field["name"])])

        if fk_ids:
            group_names = [
                f'{field["related_model_name"]}.{fk_id}.{field["related_name"]}'
                for fk_id in fk_ids
            ]
            notify_mutation(group_names, action, encoded_instance)


MODEL_RELATED_FIELDS = {}


def get_related_fields(model):
    related_fields = MODEL_RELATED_FIELDS.get(model)
    if related_fields is None:
        fields = []
        for field in model._meta.get_fields():
            if isinstance(field, (models.ForeignKey, models.ManyToManyField)):
                related_name = field.related_query_name()
                if related_name != "+":
                    is_m2m = isinstance(field, models.ManyToManyField)
                    if not is_m2m or AUTO_BROADCAST.m2m and is_m2m:
                        fields.append(
                            {
                                "is_m2m": is_m2m,
                                "name": field.attname,
                                "related_name": related_name,
                                "related_model_name": field.related_model._meta.model_name,  # noqa
                            }
                        )
        related_fields = MODEL_RELATED_FIELDS[model] = tuple(fields)
    return related_fields


@receiver(m2m_changed, is_active=AUTO_BROADCAST.m2m)
def broadcast_m2m_changed(sender, instance, action, model, pk_set, **kwargs):
    if action.startswith("post_") and instance.pk:
        encoded_instance = serializer.encode(instance)
        if action.endswith("_add"):
            action = ModelAction.ADDED
        elif action.endswith("_remove"):
            action = ModelAction.REMOVED
        elif action.endswith("_clear"):
            action = ModelAction.CLEARED
        else:
            assert False, f"Unknown action `{action}`"

        model_name = model._meta.model_name
        attr_name = get_name_of(sender, model)
        updates = [f"{model_name}.{pk}.{attr_name}" for pk in pk_set or []]
        notify_mutation(updates, action, encoded_instance)

        model = type(instance)
        model_name = model._meta.model_name
        attr_name = get_name_of(sender, model)
        update = f"{model_name}.{instance.pk}.{attr_name}"
        notify_mutation(
            [f"{update}.{pk}" for pk in pk_set or []],
            action,
            encoded_instance,
        )


def get_name_of(through, model):
    for model_field in model._meta.get_fields():
        found = getattr(model_field, "through", None) or getattr(
            getattr(model, model_field.name, None), "through", None
        )
        if through is found:
            return model_field.name


def notify_mutation(names: t.Iterable[str], action: ModelAction, instance: str):
    for name in (n.replace("_", "-") for n in names):
        log.debug(f"<-> {action} {name}")
        send_to(name, "model_mutation", action=action, instance=instance)



from django.conf import settings

from .schemas import AutoBroadcast

DEBUG = settings.DEBUG
DEFAULT = {
    "TRANSPILER_CACHE_SIZE": 1024,
    "USE_HTML_DIFF": True,
    "USE_HMIN": False,
    "BOOST_PAGES": False,
    "AUTO_BROADCAST": AutoBroadcast(),
}

REACTOR = DEFAULT | getattr(settings, "REACTOR", {})
LOGIN_URL = settings.LOGIN_URL

TRANSPILER_CACHE_SIZE: int = REACTOR["TRANSPILER_CACHE_SIZE"]
USE_HTML_DIFF: bool = REACTOR["USE_HTML_DIFF"]
USE_HMIN: bool = REACTOR["USE_HMIN"]
BOOST_PAGES: bool = REACTOR["BOOST_PAGES"]
AUTO_BROADCAST: AutoBroadcast = REACTOR["AUTO_BROADCAST"]



import json
import typing as t

from django.core.serializers.json import DjangoJSONEncoder
from lru import LRU

from .settings import TRANSPILER_CACHE_SIZE

Stack = list[t.Any]

CACHE: dict[str, str | None] = LRU(TRANSPILER_CACHE_SIZE)


def transpile(event_and_modifiers: str, command: str, kwargs: dict[str, t.Any]):
    """Translates from from the tag `on` in to JavaScript"""
    name, *modifiers = event_and_modifiers.split(".")
    cache_key = f"_handler:{modifiers}.{command}.{kwargs}"
    code: t.Optional[str] = CACHE.get(cache_key)
    if code is None:
        if not modifiers or modifiers[-1] != "inlinejs":
            modifiers.append("_reactor_code")
        code = command
        stack: Stack = [kwargs]
        while modifiers:
            modifier = modifiers.pop()
            handler: t.Optional[t.Callable[[str, Stack], str]] = getattr(
                Modifiers, modifier, None
            )
            if handler:
                code = handler(code, stack)
            else:
                stack.append(modifier)

        CACHE[cache_key] = code
    return "on" + name, code


class Modifiers:
    @staticmethod
    def _reactor_code(code: str, stack: Stack):
        kwargs = json.dumps(stack.pop(), cls=DjangoJSONEncoder)
        return f"reactor.send(event.target, '{code}', {kwargs})"

    @staticmethod
    def _add_curly(code: str):
        return "{%s}" % (code,)

    @staticmethod
    def inlinejs(code: str, stack: Stack):
        return code

    # Events

    @classmethod
    def debounce(cls, code: str, stack: Stack):
        delay = int(stack.pop())
        code = cls._add_curly(code)
        return f"reactor.debounce({delay})(() => {code})()"

    @staticmethod
    def prevent(code: str, stack: Stack):
        return "event.preventDefault(); " + code

    @staticmethod
    def stop(code: str, stack: Stack):
        return "event.stopPropagation(); " + code

    # Key modifiers

    @classmethod
    def ctrl(cls, code: str, stack: Stack):
        code = cls._add_curly(code)
        return f"if (event.ctrlKey) {code}"

    @classmethod
    def alt(cls, code: str, stack: Stack):
        code = cls._add_curly(code)
        return f"if (event.altKey) {code}"

    @classmethod
    def shift(cls, code: str, stack: Stack):
        code = cls._add_curly(code)
        return f"if (event.shiftKey) {code}"

    @classmethod
    def meta(cls, code: str, stack: Stack):
        code = cls._add_curly(code)
        return f"if (event.metaKey) {code}"

    # Key codes

    @classmethod
    def key(cls, code: str, stack: Stack):
        key = stack.pop()
        code = cls._add_curly(code)
        return f"if ((event.key + '').toLowerCase() == '{key}') {code}"

    @classmethod
    def key_code(cls, code: str, stack: Stack):
        keyCode = stack.pop()
        code = cls._add_curly(code)
        return f"if (event.keyCode == {keyCode}) {code}"

    # Key shortcuts
    @classmethod
    def enter(cls, code: str, stack: Stack):
        return cls.key(code, ["enter"])

    @classmethod
    def tab(cls, code: str, stack: Stack):
        return cls.key(code, ["tab"])

    @classmethod
    def delete(cls, code: str, stack: Stack):
        return cls.key(code, ["delete"])

    @classmethod
    def backspace(cls, code: str, stack: Stack):
        return cls.key(code, ["backspace"])

    @classmethod
    def esc(cls, code: str, stack: Stack):
        return cls.key(code, ["escape"])

    @classmethod
    def space(cls, code: str, stack: Stack):
        return cls.key(code, [" "])

    @classmethod
    def up(cls, code: str, stack: Stack):
        return cls.key(code, ["arrowup"])

    @classmethod
    def down(cls, code: str, stack: Stack):
        return cls.key(code, ["arrowdown"])

    @classmethod
    def left(cls, code: str, stack: Stack):
        return cls.key(code, ["arrowleft"])

    @classmethod
    def right(cls, code: str, stack: Stack):
        return cls.key(code, ["arrowright"])



from django.apps import AppConfig
from django.utils.module_loading import autodiscover_modules


class ReactorConfig(AppConfig):
    name = "reactor"
    verbose_name = "Django Reactor"

    def ready(self):
        from . import auto_broadcast  # noqa

        autodiscover_modules("live")



import typing as t

from django.core.serializers import deserialize, serialize
from django.core.serializers.base import DeserializedObject
from django.core.serializers.json import DjangoJSONEncoder
from django.db.models import Model
from pydantic import BaseModel

__all__ = ("encode", "decode")


def encode(instance: Model) -> str:
    return serialize("json", [instance], cls=ReactorJSONEncoder)


def decode(instance: str) -> Model:
    obj: DeserializedObject = list(deserialize("json", instance))[0]
    obj.object.save = obj.save
    return obj.object


class ReactorJSONEncoder(DjangoJSONEncoder):
    def default(self, o: t.Any) -> t.Any:
        if isinstance(o, BaseModel):
            return o.dict()
        else:
            return super().default(o)



import json
import logging
import typing as t

from channels.generic.websocket import AsyncJsonWebsocketConsumer
from django.contrib.auth.models import AnonymousUser
from django.core.signing import Signer
from django.utils.datastructures import MultiValueDict
from reactor.component import Component

from . import serializer
from .repository import ComponentRepository
from .utils import parse_request_data

log = logging.getLogger("reactor")


class ReactorConsumer(AsyncJsonWebsocketConsumer):
    @property
    def user(self):
        return self.scope.get("user") or AnonymousUser()

    async def connect(self):
        await super().connect()
        self.subscriptions = set()
        self.query_stirng: str = ""
        self.repo = ComponentRepository(
            user=self.user,
            channel_name=self.channel_name,
            channel_layer=self.channel_layer,
        )

    # Fronted commands

    async def receive_json(self, content):
        await getattr(self, f'command_{content["command"]}')(
            **content["payload"]
        )

    async def command_join(
        self,
        name: str,
        state: str,
        parent_id: str | None = None,
    ):
        decoded_state: dict[str, t.Any] = json.loads(Signer().unsign(state))
        log.debug(f"<<< JOIN {name} {decoded_state}")
        try:
            component, created = await self.repo.join(
                name, decoded_state, parent_id=parent_id
            )
        except Exception as e:
            log.exception(e)
            if id := decoded_state.get("id"):
                await self.component_remove(id)
        else:
            if created:
                await self.send_render(component)
            await self.after_mutation_schores()

    async def command_leave(self, id):
        log.debug(f"<<< LEAVE {id}")
        self.repo.remove(id)

    async def command_query_string(self, qs: str):
        self.query_string = qs
        self.repo.set_query_string(qs)

    async def command_user_event(
        self, id, command, implicit_args, explicit_args
    ):
        kwargs = dict(
            parse_request_data(MultiValueDict(implicit_args)), **explicit_args
        )
        log.debug(f"<<< USER-EVENT {id} {command} {kwargs}")
        component = await self.repo.dispatch_event(id, command, kwargs)
        await self.send_render(component)
        await self.after_mutation_schores()

    # Component commands

    async def message_from_component(self, data):
        await getattr(self, f"component_{data['command']}")(**data["kwargs"])

    async def component_remove(self, id):
        log.debug(f">>> REMOVE {id}")
        await self.send_command("remove", {"id": id})

    async def component_send_render(self, id):
        log.debug(f">>> SEND-RENDER {id}")
        if component := self.repo.get(id):
            await self.send_render(component)

    async def component_dom_action(self, action, id, html):
        log.debug(f">>> DOM {action.upper()} {id}")
        await self.send_command(action, {"id": id, "html": html})

    async def component_scroll_into_view(self, id, behavoir, block, inline):
        log.debug(f">>> SCROLL-INTO-VIEW {id}")
        await self.send_command(
            "scroll_into_view",
            {"id": id, "behavoir": behavoir, "block": block, "inline": inline},
        )

    async def component_focus_on(self, selector):
        log.debug(f'>>> FOCUS ON "{selector}"')
        await self.send_command("focus_on", {"selector": selector})

    async def component_url_change(self, command: str, url: str):
        log.debug(f'>>> URL {command.upper()} "{url}"')
        await self.send_command("url_change", {"url": url, "command": command})

    # Incoming messages from subscriptions

    async def model_mutation(self, data):
        # The signature here is coupled to:
        #   `reactor.auto_broadcast.notify_mutation`
        await self._dispatch_notifications(
            "mutation",
            data["channel"],
            {
                "instance": serializer.decode(data["instance"]),
                "action": data["action"],
            },
        )

    async def notification(self, data):
        # The signature here is coupled to:
        #   `reactor.utils.send_notification`
        await self._dispatch_notifications(
            "notification", data["channel"], data["kwargs"]
        )

    async def _dispatch_notifications(
        self, receiver: str, channel: str, kwargs: dict[str, t.Any]
    ):
        for component in self.repo.components_subscribed_to(channel):
            await getattr(component, receiver)(channel, **kwargs)
            await self.send_render(component)
        await self.after_mutation_schores()

    # Reply to front-end

    async def send_render(self, component: Component):
        diff = await component._render_diff(self.repo)
        if diff is not None:
            log.debug(f">>> RENDER {component._name} {component.id}")
            await self.send_command(
                "render",
                {"id": component.id, "diff": diff},
            )

    async def send_command(self, command, payload):
        await self.send_json({"command": command, "payload": payload})

    async def after_mutation_schores(self):
        await self.update_to_which_channels_im_subscribed_to()
        await self.send_query_string()

    async def update_to_which_channels_im_subscribed_to(self):
        if self.channel_layer is not None and self.channel_name is not None:
            subscriptions = self.repo.subscriptions

            # new subscriptions
            for channel in subscriptions - self.subscriptions:
                log.debug(f"::: SUBSCRIBE {self.channel_name} to {channel}")
                await self.channel_layer.group_add(channel, self.channel_name)

            # remove subscriptions
            for channel in self.subscriptions - subscriptions:
                log.debug(f"::: UNSUBSCRIBE {self.channel_name} to {channel}")
                await self.channel_layer.group_discard(
                    channel, self.channel_name
                )

            self.subscriptions = subscriptions

    async def send_query_string(self):
        new_qs = self.repo.get_query_string()
        if self.query_string != new_qs:
            self.query_string = new_qs
            log.debug(f">>> QS {new_qs}")
            await self.send_command("set_query_string", {"qs": new_qs})



import difflib
import typing as t
from asyncio import iscoroutine, iscoroutinefunction
from functools import reduce
from uuid import uuid4

from asgiref.sync import async_to_sync
from channels.db import database_sync_to_async as db
from channels.layers import BaseChannelLayer
from django.apps import apps
from django.contrib.auth.base_user import AbstractBaseUser
from django.contrib.auth.models import AnonymousUser
from django.db import models
from django.http import HttpRequest
from django.shortcuts import resolve_url  # type: ignore
from django.template import loader
from django.utils.html import format_html
from django.utils.safestring import SafeString, SafeText, mark_safe
from pydantic import BaseModel, validate_arguments
from pydantic.fields import Field, ModelField

from . import settings, utils
from .schemas import DomAction, ModelAction

if settings.USE_HMIN:
    try:
        from hmin.base import html_minify  # type: ignore
    except ImportError as e:
        raise ImportError(
            "If you enable REACTOR['USE_HMIN'] you need to install django-hmin"
        ) from e
else:

    def html_minify(html: str) -> str:
        return html


ComponentState = Context = MessagePayload = dict[str, t.Any]
RedirectDestination = t.Callable[(...), t.Any] | models.Model | str
HTMLDiff = list[str | int]
ComponentOrHtml = t.Union["Component", SafeString]
P = t.ParamSpec("P")


class Template(t.Protocol):
    def render(
        self,
        context: Context | dict[str, t.Any] | None = ...,
        request: HttpRequest | None = ...,
    ) -> SafeString:
        ...


class Repo(t.Protocol):
    pass


ScrollPosition = (
    t.Literal["start"]
    | t.Literal["end"]
    | t.Literal["center"]
    | t.Literal["nearest"]
)


__all__ = ("Component", "broadcast")


def broadcast(channel: str, **kwargs: t.Any):
    utils.send_to(channel, type="notification", kwargs=kwargs)


class ReactorMeta:
    _last_sent_html: list[str]

    def __init__(
        self,
        params: dict[str, t.Any],
        channel_name: str | None = None,
        channel_layer: BaseChannelLayer | None = None,
        parent_id: str | None = None,
    ):
        self.params = params
        self.channel_name = channel_name
        self.channel_layer = channel_layer
        self.parent_id = parent_id
        self._is_frozen: bool = False
        self._redirected_to: str | None = None
        self._last_sent_html: list[str] = []
        self._skip_render: bool = False

    def clone(self):
        return type(self)(
            params=self.params,
            channel_name=self.channel_name,
            channel_layer=self.channel_layer,
        )

    def skip_render(self):
        self._skip_render = True

    def force_render(self):
        self._skip_render = False
        self._last_sent_html = []

    async def destroy(self, component_id: str):
        self.freeze()
        await self.send("remove", id=component_id)

    def freeze(self):
        self._is_frozen = True

    async def redirect_to(self, to: RedirectDestination, **kwargs: t.Any):
        url = resolve_url(to, **kwargs)
        self._redirected_to = url
        if self.channel_name:
            self.freeze()
            await self.send("url_change", command="redirect", url=url)

    async def replace_to(self, to: RedirectDestination, **kwargs):
        url = resolve_url(to, **kwargs)
        await self.send("url_change", command="replace", url=url)

    async def push_to(self, to: RedirectDestination, **kwargs):
        url = resolve_url(to, **kwargs)
        await self.send("url_change", command="push", url=url)

    async def render_diff(
        self, component: "Component", repo: Repo
    ) -> HTMLDiff | None:
        if self._skip_render:
            self._skip_render = False
        else:
            html = await db(self.render)(component, repo)
            if html and self._last_sent_html != (html := html.split(" ")):
                if settings.USE_HTML_DIFF:
                    diff: HTMLDiff = []
                    for x in difflib.ndiff(self._last_sent_html, html):
                        indicator = x[0]
                        if indicator == " ":
                            diff.append(1)
                        elif indicator == "+":
                            diff.append(x[2:])
                        elif indicator == "-":
                            diff.append(-1)

                    if diff:
                        diff = reduce(compress_diff, diff[1:], diff[:1])
                else:
                    diff = html  # type: ignore
                self._last_sent_html = html
                return diff

    def render(self, component: "Component", repo: Repo) -> None | SafeText:
        html = None
        if not self.channel_name and self._redirected_to:
            html = format_html(
                '<meta http-equiv="refresh" content="0; url={url}">',
                url=self._redirected_to,
            )
        elif not (self._is_frozen or self._redirected_to) and html is None:
            template = component._get_template()
            context = self._get_context(component, repo)
            html = template.render(context).strip()
            html = html_minify(html)
        if html:
            return mark_safe(html)

    async def send_dom_action(
        self,
        action: DomAction,
        id: str,
        html: SafeString,
    ):
        await self.send("dom_action", action=action.value, id=id, html=html)

    async def scroll_into_view(
        self,
        id: str,
        behavoir: t.Literal["smooth"]
        | t.Literal["instant"]
        | t.Literal["auto"] = "auto",
        block: ScrollPosition = "start",
        inline: ScrollPosition = "nearest",
    ):
        await self.send(
            "scroll_into_view",
            id=id,
            behavoir=behavoir,
            block=block,
            inline=inline,
        )

    async def send(self, _command: str, **kwargs: t.Any):
        if self.channel_name:
            await self.send_to(self.channel_name, _command, **kwargs)

    async def send_to(self, _channel: str, _command: str, **kwargs: t.Any):
        if self.channel_layer:
            await self.channel_layer.send(
                _channel,
                {
                    "type": "message_from_component",
                    "command": _command,
                    "kwargs": kwargs,
                },
            )

    def _get_context(
        self,
        component: "Component",
        repo: Repo,
    ) -> Context:
        context = {}

        for attr_name in dir(component):
            if not attr_name.startswith("_"):
                attr = getattr(component, attr_name)
                if not callable(attr):
                    if iscoroutine(attr) or iscoroutinefunction(attr):
                        attr = async_to_sync(attr)
                    context[attr_name] = attr
        return dict(
            context,
            this=component,
            reactor_repository=repo,
        )


def compress_diff(diff: HTMLDiff, diff_item: str | int) -> HTMLDiff:
    if isinstance(diff_item, str) or isinstance(diff[-1], str):
        diff.append(diff_item)
    else:
        same_sign = not (diff[-1] > 0) ^ (diff_item > 0)
        if same_sign:
            diff[-1] += diff_item
        else:
            diff.append(diff_item)
    return diff


def load_model_instance(model, v, fields, field: ModelField, config):
    if v is None or isinstance(v, field.type_):
        return v
    else:
        return field.type_.objects.filter(pk=v).first()


def load_queryset(model, v, fields, field: ModelField, config):
    if isinstance(v, field.type_):
        return v
    else:
        return apps.get_model(v["app"], v["model"]).objects.filter(  # type: ignore
            pk__in=v["ids"]
        )


class Component(BaseModel):
    __name__: str

    _all: dict[str, t.Type["Component"]] = {}
    _urls = {}
    _name: str = ...  # type: ignore
    _template_name: str = ...  # type: ignore
    _templates: dict[str, Template] = {}
    _fqn: str
    _tag_name: str

    # HTML tag that this component extends
    _extends = "div"

    # fields to exclude from the component state during serialization
    _exclude_fields = {"user", "reactor"}

    # Subscriptions: you can define here which channels this component is
    # subscribed to
    _subscriptions: set[str] = set()

    class Config:
        arbitrary_types_allowed = True
        validate_assignment = True
        json_encoders = {
            models.Model: lambda x: x.pk,
            models.QuerySet: lambda qs: {
                "app": qs.model._meta.app_label,
                "model": qs.model._meta.model_name,
                "ids": [x.pk for x in qs],
            },
        }

    def __init_subclass__(
        cls: t.Type["Component"], name: str | None = None, public: bool = True
    ):
        if public:
            name = name or cls.__name__
            cls._all[name] = cls
            # Component name
            cls._name = name
            # Fully qualified name
            cls._fqn = f"{cls.__module__}.{name}"

            # Compote a valid HTML tag from the component name
            name = "".join([("-" + c if c.isupper() else c) for c in name])
            name = name.strip("-").lower()
            cls._tag_name = "x-" + name

        for attr_name in vars(cls):
            attr = getattr(cls, attr_name)
            if (
                not attr_name.startswith("_")
                and attr_name.islower()
                and callable(attr)
            ):
                setattr(
                    cls,
                    attr_name,
                    validate_arguments(
                        config={"arbitrary_types_allowed": True}
                    )(attr),
                )

        # Hook up the Model loaders
        for field in cls.__fields__.values():
            if field.pre_validators is None:
                try:
                    is_model = issubclass(field.type_, models.Model)
                    is_qs = issubclass(field.type_, models.QuerySet)  # type: ignore
                except TypeError:
                    is_model = False
                    is_qs = False

                if is_model:
                    field.pre_validators = [load_model_instance]
                elif is_qs:
                    field.pre_validators = [load_queryset]

        super().__init_subclass__()

    @classmethod
    def _build(
        cls,
        _component_name: str,
        state: ComponentState,
        params: dict[str, t.Any],
        user: AnonymousUser | AbstractBaseUser | None = None,
        channel_name: str | None = None,
        channel_layer: BaseChannelLayer | None = None,
        parent_id: str | None = None,
    ) -> "Component":
        if _component_name not in cls._all:
            raise ComponentNotFound(
                (
                    f"Could not find requested component '{_component_name}'. "
                    f"Did you load the component?"
                )
            )

        # TODO: rename state to initial_state
        instance = cls._all[_component_name].new(
            user=user or AnonymousUser(),
            reactor=ReactorMeta(
                params=params,
                channel_name=channel_name,
                channel_layer=channel_layer,
                parent_id=parent_id,
            ),
            **state,
        )
        return instance

    @classmethod
    def _get_template(cls, template_name: str | None = None) -> Template:
        template_name = template_name or cls._template_name
        if settings.DEBUG:
            return loader.get_template(template_name)
        else:
            if (template := cls._templates.get(template_name)) is None:
                template = loader.get_template(template_name)
                cls._templates[template_name] = template
            return template

    # State
    id: str = Field(default_factory=lambda: f"rx-{uuid4()}")
    user: AnonymousUser | AbstractBaseUser
    reactor: ReactorMeta

    @classmethod
    def new(cls, **kwargs: t.Any):
        return cls(**kwargs)

    async def joined(self):
        ...

    async def mutation(
        self, channel: str, action: ModelAction, instance: t.Any
    ):
        ...

    async def notification(self, channel: str, **kwargs: t.Any):
        ...

    async def destroy(self):
        await self.reactor.destroy(self.id)

    async def send_render(self):
        await self.reactor.send("send_render", id=self.id)

    async def focus_on(self, selector: str):
        await self.reactor.send("focus_on", selector=selector)

    # Dom operations

    def skip_render(self):
        self.reactor.skip_render()

    def force_render(self):
        self.reactor.force_render()

    async def dom(
        self,
        _action: DomAction,
        _id: str,
        _component_class_or_template_name: t.Type["Component"] | str,
        **kwargs,
    ):
        if isinstance(_component_class_or_template_name, str):
            template = self._get_template(_component_class_or_template_name)
            html = await db(template.render)(kwargs)
        else:
            from .repository import ComponentRepository

            component = _component_class_or_template_name.new(
                reactor=self.reactor.clone(),
                user=self.user,
                **kwargs,
            )
            html = await db(component._render)(
                ComponentRepository(user=self.user, params=self.reactor.params)
            )
        await self.reactor.send_dom_action(_action, _id, html)

    # Internal render operations

    def _render(self, repo: Repo):
        return self.reactor.render(self, repo)

    def _render_diff(self, repo: Repo):
        return self.reactor.render_diff(self, repo)


class ComponentNotFound(LookupError):
    pass



import inspect
import logging
import typing as t
from collections import defaultdict
from functools import wraps

from asgiref.sync import async_to_sync
from channels.layers import get_channel_layer
from django.utils.datastructures import MultiValueDict

log = logging.getLogger("reactor")


def on_commit(f: t.Callable[(...), None]):
    @wraps(f)
    def wrapper(*args: t.Any, **kwargs: t.Any):
        from django.db.transaction import on_commit  # type: ignore

        on_commit(lambda: f(*args, **kwargs))

    return wrapper


@on_commit
def send_to(channel: t.Optional[str], type: str, **kwargs: t.Any):
    """Sends a message of `type` to the"""
    if channel:
        async_to_sync(get_channel_layer().group_send)(
            channel, dict(type=type, channel=channel, **kwargs)
        )


@on_commit
def send_notification(channel: str, **kwargs):
    log.debug(f"<-> NOTIFICATION {channel} {kwargs}")
    send_to(channel, "notification", kwargs=kwargs)


# Introspection


def filter_parameters(f, kwargs):
    has_kwargs = any(
        param.kind == inspect.Parameter.VAR_KEYWORD
        for param in inspect.signature(f).parameters.values()
    )
    if has_kwargs:
        return kwargs
    else:
        return {
            param: value
            for param, value in kwargs.items()
            if param in f.model.__fields__
        }


# Decoder for client requests
def parse_request_data(data: MultiValueDict[str, t.Any]):
    return _parse_obj(_extract_data(data))


def _extract_data(data: MultiValueDict[str, t.Any]):
    for key in set(data):
        if key.endswith("[]"):
            key = key.removesuffix("[]")
            value = data.getlist(key)
        else:
            value = data.get(key)
        yield key.split("."), value


def _parse_obj(
    data: t.Iterable[tuple[list[str], t.Any]], output=None
) -> dict[str, t.Any] | t.Any:
    output = output or {}
    arrays = defaultdict(lambda: defaultdict(dict))  # field -> index -> value
    for key, value in data:
        fragment, *tail = key
        if "[" in fragment:
            field_name = fragment[: fragment.index("[")]
            index = int(fragment[fragment.index("[") + 1 : -1])
            arrays[field_name][index] = (
                _parse_obj([(tail, value)], arrays[field_name][index])
                if tail
                else value
            )
        else:
            output[fragment] = _parse_obj([(tail, value)]) if tail else value

    for field, items in arrays.items():
        output[field] = [
            v for _, v in sorted(items.items(), key=lambda kv: kv[0])
        ]
    return output


data = [
    (["a"], [2, 2, 3, 5, 5]),
    (["b"], 2),
    (["x[0]"], 10),
    (["x[1]"], 20),
    (["c[1]", "a"], 1),
    (["c[0]", "a"], 3),
    (["c[1]", "b"], 1),
    (["c[0]", "b"], 4),
]


assert _parse_obj(data) == {
    "a": [2, 2, 3, 5, 5],
    "b": 2,
    "c": [{"a": 3, "b": 4}, {"a": 1, "b": 1}],
    "x": [10, 20],
}



import typing as t

from django import template
from django.core.signing import Signer
from django.template.base import Node, Parser, Token
from django.utils.html import format_html

from .. import settings
from ..component import Component
from ..event_transpiler import transpile
from ..repository import ComponentRepository

register = template.Library()


@register.inclusion_tag("reactor_header.html")
def reactor_header():
    return {
        "components": [
            {"tag_name": component._tag_name, "extends": component._extends}
            for component in Component._all.values()
        ],
        "BOOST_PAGES": settings.BOOST_PAGES,
    }


@register.simple_tag(takes_context=True)
def tag_header(context):
    component: Component = context["this"]
    return format_html(
        (
            'is="{tag_name}" '
            'id="{id}" '
            'data-name="{name}" '
            'data-state="{state}" '
            "reactor-component"
        ),
        tag_name=component._tag_name,
        id=component.id,
        name=component._name,
        state=Signer().sign(component.json(exclude=component._exclude_fields)),
    )


@register.simple_tag(takes_context=True)
def component(context, _name, **kwargs):
    parent: t.Optional[Component] = context.get("this")
    parent_id = parent.id if parent else None

    if (repo := context.get("reactor_repository")) is None:
        qs = (
            (request := context.get("request"))
            and request.META["QUERY_STRING"]
            or ""
        )
        repo = ComponentRepository(
            user=context.get("user"),
            params=ComponentRepository.extract_params(qs),
        )

    component, _created = repo.build(_name, state=kwargs, parent_id=parent_id)
    return component._render(repo) or ""


@register.simple_tag(takes_context=True)
def on(context, _event_and_modifiers, _command, **kwargs: t.Any):
    component: t.Optional[Component] = context.get("this")

    assert component, "Can't find a component in this context"
    handler = getattr(component, _command, None)
    assert handler, f"Missing handler: {component._name}.{_command}"
    assert callable(handler), f"Not callable: {component._name}.{_command}"

    event, code = transpile(_event_and_modifiers, _command, kwargs)
    return format_html('{event}="{code}"', event=event, code=code)


@register.filter(name="str")
def to_string(value):
    return str(value)


@register.filter
def concat(value, arg):
    return f"{value}{arg}"


# Shortcuts and helpers


@register.tag()
def cond(parser: Parser, token: Token):
    """Prints some text conditionally

        ```html
        {% cond {'works': True, 'does not work': 1 == 2} %}
        ```
    Will output 'works'.
    """
    dict_expression = token.contents[len("cond ") :]
    return CondNode(dict_expression)


@register.tag(name="class")
def class_cond(parser: Parser, token: Token):
    """Prints classes conditionally

    ```html
    <div {% class {'btn': True, 'loading': loading, 'falsy': 0} %}></div>
    ```

    If `loading` is `True` will print:

    ```html
    <div class="btn loading"></div>
    ```
    """
    dict_expression = token.contents[len("class ") :]
    return ClassNode(dict_expression)


class CondNode(Node):
    def __init__(self, dict_expression):
        self.dict_expression = dict_expression

    def render(self, context):
        variables: dict[str, t.Any] = context.flatten()  # type: ignore
        terms = eval(self.dict_expression, variables)
        return " ".join(term for term, ok in terms.items() if ok)


class ClassNode(CondNode):
    def render(self, *args, **kwargs):
        text = super().render(*args, **kwargs)
        return f'class="{text}"'



from enum import StrEnum

from pydantic import BaseModel, Field


class AutoBroadcast(BaseModel):
    # model-a
    model: bool = False
    # model-a.1234
    model_pk: bool = False
    # model-b.9876.model-a-set
    related: bool = False
    # model-b.9876.model-a-set
    # model-a.1234.model-b-set
    m2m: bool = False
    # this is a set of tuples of ('app_label', 'ModelName')
    # to subscribe for the auto broadcast
    senders: set[tuple[str, str]] = Field(default_factory=set)


class ModelAction(StrEnum):
    # Model action
    UPDATED = "UPDATED"
    DELETED = "DELETED"
    CREATED = "CREATED"

    # M2M actions
    ADDED = "ADDED"
    REMOVED = "REMOVED"
    CLEARED = "CLEARED"


class DomAction(StrEnum):
    APPEND = "append"
    PREPEND = "prepend"
    INSERT_AFTER = "insert_after"
    INSERT_BEFORE = "insert_before"
    REPLACE_WITH = "replace_with"
